/*
  Vt1xx / ANSI terminal emulation by
  Rene Rebe <rene@exactcode.de>
  
  Lossy based on code of bogl-term by:
  Edmund GRIMLEY EVANS <edmundo@rano.org>
  
  Additional information from the Linux kernel driver/char/vt.c by:
  Linus Torvalds
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA. */

#include <stdlib.h>
#include <string.h> // memset

#include <iostream>

#include "Term.hh"

#define SCR(x, y) \
  ((x) + (((y) /*+ yorig*/) % ysize) * xsize)

Term::Term (int _xsize, int _ysize)
{
  xpos = ypos = saved_xpos = saved_ypos = 0;
  xp = yp = 0; // silence Valgrind
  saved = false;
  
  fg = def_fg = 0;
  bg = def_bg = 7;
  
  rev = 0;
  state = 0;
  cur_visible = 1;
  memset (&ps, 0, sizeof(&ps));

  screen   = 0;
  screenfg = 0;
  screenbg = 0;
  screenul = 0;
  
  xsize = ysize = 0;
  resize (_xsize, _ysize);
}

void Term::resize (int _xsize, int _ysize)
{
  if (xsize == _xsize && ysize == _ysize)
    return;
  
  if (screen || screenfg || screenbg || screenul)
    {
      // TODO: copy content ...
      free (screen);
      free (screenfg);
      free (screenbg);
      free (screenul);
    }
  
  xsize = _xsize;
  ysize = _ysize;;
  scrollA = 0;
  scrollB = ysize - 1;
  
  screen   = (wchar_t*) malloc(xsize * ysize * sizeof(wchar_t));
  screenfg = (int*) malloc(xsize * ysize * sizeof(int));
  screenbg = (int*) malloc(xsize * ysize * sizeof(int));
  screenul = (int*) malloc(xsize * ysize * sizeof(int));

  for (int i = 0; i < xsize * ysize; i++) {
    screen[i] = ' ';
    screenfg[i] = def_fg;
    screenbg[i] = def_bg;
    screenul[i] = 0;
  }
}

bool Term::is_clear (int p1)
{
    if(screen[p1] != ' ')
    	return false;
    if(screenfg[p1] != fg)
    	return false;
    if(screenbg[p1] != bg)
    	return false;
    if(screenul[p1] != 0)
    	return false;
    return true;
}

void Term::cursor_down ()
{
    if (ypos < scrollB) {
        ++ypos;
        return;
    }

    // scroll one line
    scrolldown ();
    
    for (int i = 0; i < xsize; ++i)
      {
        int p = SCR(i, ypos);

        screen[p] = ' ';
        screenfg[p] = fg;
        screenbg[p] = bg;
        screenul[p] = 0;
    }
}

void Term::scrolldown ()
{
  for (int y = scrollA+1; y <= scrollB; ++y)
    {
      int this_point = SCR(0,y);
      int next_point = SCR(0,y-1);
      for (int x = 0; x < xsize; ++x)
	{
	  screen   [next_point] = screen   [this_point];
	  screenfg [next_point] = screenfg [this_point];
	  screenbg [next_point] = screenbg [this_point];
	  screenul [next_point] = screenul [this_point];
	  
	  ++this_point; ++next_point;
	}
    }
  // :-(
  redraw ();
}

void Term::scrollup ()
{
  /* Scroll up one line.  */
  //  if (ypos > 0)
  //    ypos--;
  //  else
  //   {
  
  ypos = 0;
  
  for (int y = scrollB-1; y >= scrollA; --y)
    {
      int this_point = SCR(0,y);
      int next_point = SCR(0,y+1);
      for (int x = 0; x < xsize; ++x)
	{
	  screen   [next_point] = screen   [this_point];
	  screenfg [next_point] = screenfg [this_point];
	  screenbg [next_point] = screenbg [this_point];
	  screenul [next_point] = screenul [this_point];
	  
	  ++this_point; ++next_point;
	}
    }
  
  /* Clear the top line.  */
  for (int i = SCR (0, scrollA); i < SCR (xsize, scrollA); ++i)
    {
      screen[i] = ' ';
      screenfg[i] = fg;
      screenbg[i] = bg;
      screenul[i] = 0;
    }
}

void Term::show_cursor (bool show)
{
    int i, x, fg, bg;

    if ((x = xpos) == xsize)
        x = xsize - 1;

    i = SCR(x, ypos);

    while (!screen[i])
        --i, --x;

    if (screen[i])
    {
        if ((show && !rev) || (!show && rev))
            fg = screenbg[i], bg = screenfg[i];
        else
            fg = screenfg[i], bg = screenbg[i];
        put_char (x, ypos, screen[i], fg, bg, screenul[i]);
    }
}

void Term::clear_left ()
{
  int i = SCR(0, ypos);
  for (int x = 0; x < xpos; ++x)
    clear_one (i++);
}

void Term::clear_right ()
{
  int i = SCR(xpos, ypos);
  for (int x = xpos; x < xsize; ++x)
    clear_one (i++);
}

void Term::clear_one (int i)
{
  if(!is_clear (i))
    {
      screen[i] = ' ';
      screenfg[i] = fg;
      screenbg[i] = bg;
      screenul[i] = 0;
    }
}

void Term::clear_line (int y)
{
  int i = SCR(0, y);

  for (int x = 0; x < xsize; ++x, ++i)
  {
    if (screen[i] != ' ')
    	screen[i] = ' ';
  }
}

void Term::save ()
{
  saved = true;
  saved_xpos = xpos;
  saved_ypos = ypos;
}

void Term::unsave ()
{
  if (saved) {
    saved_xpos = xpos;
    saved_ypos = ypos;
  } else {
    std::cerr << "unsave without save!" << std::endl;
  }
}

void Term::out (char *s, int n)
{
    int i,w, txp, f, b, use_acs;
    size_t k, kk;
    wchar_t wc;
    
    char buf [MB_LEN_MAX];
    
    const bool debug = false;

    if (debug) {
      for (int i = 0; i < n; ++i) {
	//switch (s[i]) {
	//default:
	if (isprint (s[i])) 
	  fprintf (stderr, "%c", s[i]);
	else {
	  if (s[i] == 27)
	    fprintf (stderr, "\nESC");
	  else
	    fprintf (stderr, " \\0%o ", (int) ((unsigned char) s[i]) );
	}
	fflush (stderr);
      }
    }

    k = 0;
    while (1)
    {
	s += k;
	n -= k;

	/* The n <= 0 check was originally only necessary because of a bug
	   (?) in glibc 2.2.3, as opposed to libiconv.  glibc will
	   successfully convert a zero-length string.  It is also the only
	   exit point from this loop when we run out of characters, whether
	   we successfully decode a zero-length string or error out.  The
	   exception is an incomplete multibyte sequence, just below.  */
	if (n <= 0)
	    break;

	k = mbrtowc (&wc, s, n, &ps);
	
	//std::cerr << "mbrtowc: " << (int) k <<  std::endl;
	
	/* If we fail to write a character, skip forward one byte and continue.
	   There's not much we can do to recover, but it's better than discarding
	   the whole line.  */
	if (k == -1)
	{
	    k = 1;
	    /* The mbrtowc documentation suggests that we could use mbrtowc
	       to reset ps, but that doesn't work in practice; ps is in
	       an undefined state which appears to be the illegal state to make
	       the reset call in.  Use memset.  */
	    memset (&ps, 0, sizeof (ps));
	    if (debug)
	      std::cerr << "mbrtowc failed" << std::endl;
	    clear_one (SCR(xpos,ypos));xpos++; // HACK - TODO: rewrite parser
	    continue;
	}
	else if (k == -2)
	{
	    /* Incomplete character, so we exit and wait for more to arrive.  */
	    break;
	}

        if (!k)
            k = 1;

        txp = xp;
        xp = -1;

        if (wc == 0)            /* 0 has a special meaning in screen[] */
            continue;
	
        if (wc == 7)            /* TODO: bell :-) */
	  { 
	    state = 0;
	    continue;
	  }
	
        if (wc == 8)
        {                       /* cub1=^H */
            if (xpos)
                --xpos;
            state = 0;
            continue;
        }

        if (wc == 9) // Tab
        {                       /* ht=^I */
	    /* I'm not sure whether this way of going over the right margin
               is correct, so I don't declare this capability in terminfo. */
            xpos = (xpos / 8) * 8 + 8;
	    // the oeig code wiped the movement with ' ' wich is
	    // not what nvi et al. expects
            state = 0;
	    continue;
        }

        if (wc == 10)
        {                       /* ind=^J */
            cursor_down ();
            state = 0;
            continue;
        }

        if (wc == 13)
        {                       /* cr=^M */
            xpos = 0;
            state = 0;
            continue;
        }

	if (wc == 14)
	{
	    acs = 1;
	    continue;
	}

	if (wc == 15)
	{
	    acs = 0;
	    continue;
	}

        if (wc == 27)
        {                       /* ESC = \E */
            state = -1;
            continue;
        }
	
	// vim color scroll
	// [?25l - cursor off
	// [?1c
	// [1;24r
	// [0m - reset attribute
	// [24;1H
	// [1;25r
	// [24;1H
	// [1m - set bright
	// [34m - set blue
	// [0m - reset attribute
	// [1m - set bright
	// [35m - set magenta
	// [?25h 
	// [?0c
	
	// vim scroll
	// [?25l - cursor control
	// [?1c
	// [1;24r - enable scrolling from row 1 to 24
	// [24;1H - set cursor home 24:1
	// [1;25r - enable scrolling from row 1 to 25
	// [24;1H - set cursor home 24:1
	// [?25h
	// [?0c

        if (state == -1)
        {
	  // std::cerr << "state -1: " << (char)wc << std::endl;
	  switch (wc)
	    {
	      
	    case '7':
	      save();
	      state = 0;
	      break;
		
	    case '8':
	      unsave();
	      state = 0;
	      break;
	      
	    case '[':
	      arg[0] = 0;
	      state = 1;
	      break;

	      // TODO: case ']': ?
	      /*case 'R': // reset palette
		state = 0;
		break;
	      */
	    case 'M':
	      scrollup();
	      state = 0;
	      break;
	    
	    case 'D':
	      /* Scroll down one line.  */
	      scrolldown ();
	      state = 0;
	      break;
	      
	      // e.g. reset
	      // 'H'
	      // 'c'
	      // ']'
	      // 'R'
	      
	    default:
	      std::cerr << "Unknown code: '" << (char)wc
			<< "' (#" << (int)wc
			<< " - state " << state << ")" << std::endl;

	    }
	  continue;
        }

        if (state > 0)
        {
	  //std::cerr << "state >0: " << (char)wc << std::endl;
	  
	  switch (wc)
	    {
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	      if (state > sizeof (arg) / sizeof (int))
		continue;
	      
	      arg[state - 1] *= 10;
	      arg[state - 1] += wc - '0';
	      break;
	    
	    case ';':
	      if (state > sizeof (arg) / sizeof (int))
		continue;
	      if (state < sizeof (arg) / sizeof (int))
		arg[state] = 0;
	      
	      ++state;
	      break;
	      
	    case '?':
	      if (state == 1 && arg[0] == 0)
		++state, arg[0] = -1, arg[1] = 0;
	      else
		state = 0;
	      break;
	      
	    case 'H':
	      /* home=\E[H, cup=\E[%i%p1%d;%p2%dH */
	      if (state < 3)
                {
		  if (state == 2)
                    {
		      if (arg[1] <= xsize)
			xpos = arg[1] ? arg[1] - 1 : 0;
                    }
		  else
		    xpos = 0;
		  if (arg[0] <= ysize)
		    ypos = arg[0] ? arg[0] - 1 : 0;
		}
	      state = 0;
	      break;
	      
	    case 'J': // Erase in display
	      /* clear=\E[H\E[2J */
	      /* arg[0]: 0 means clear cursor to end, 1 should mean clear
		 until cursor, 2 means clear whole screen. */
	      if (state == 1 && arg[0] == 2) // whole
		{
		  for (int i = 0; i < ysize; ++i)
		    clear_line (i);
		}
	      else if (state == 1 && arg[0] == 1) // until cursor
		{
		  for (int i = 0; i < ypos; ++i)
		    clear_line (i);
		  clear_left ();
		}
	      else // no arg or explicit 0 - cursor to end
		{
		  clear_right ();
		  for (int i = ypos+1; i < ysize; ++i)
		    clear_line (i);
		}
	      state = 0;
	      break;
	      
	    case 'K': // Erase in line
	      /* el=\E[K */
	      /* arg[0]: 0 means clear cursor to end, 1 should mean clear
		 until cursor, 2 means clear whole line. */
	      if (state == 1 && arg[0] == 2) // whole
		clear_line (ypos);
	      else if (state == 1 && arg[0] == 1) // until cursor
		clear_left ();
	      else // no arg or explicit - cursor to end
		clear_right ();
	      state = 0;
	      break;
	      
	    case 'h':
	      /* cnorm=\E[?25h */
	      if (arg[0] == -1 && arg[1] == 25)
		cur_visible = 1;
	      state = 0;
	      break;
	      
	    case 'l':
	      /* civis=\E[?25l */
	      if (arg[0] == -1 && arg[1] == 25)
		cur_visible = 0;
	      
	      state = 0;
	      break;
	      
	    case 'm':
	      /* setab=\E[4%p1%dm, setaf=\E[3%p1%dm */
	      /* case 5: blink */
	     
	      if (arg[0] == 4 || arg[0] == 24)
		ul = arg[0] == 4;
	      else if (30 <= arg[0] && arg[0] < 38)
		fg = arg[0] - 30;
	      else if (40 <= arg[0] && arg[0] < 48)
		bg = arg[0] - 40;
	      else if (arg[0] == 39)
		fg = def_fg;
	      else if (arg[0] == 49)
		bg = def_bg;
	      else if (arg[0] == 7)
		rev = 1;
	      else if (arg[0] == 27)
		rev = 0;
	      else if (arg[0] == 0)
                {
		  rev = 0;
		  fg = def_fg;
		  bg = def_bg;
                }
	      
	      state = 0;
	      break;
	      
	    case 'r':
	      // set scroll range
	      if (state == 2)
		{
		  scrollA = arg[0] - 1;
		  scrollB = arg[1] - 1;
		  if (scrollB >= ysize)
		    scrollB = ysize;
		  if (scrollA >= scrollB)
		    scrollA = scrollB - 1;
		}
	      state = 0;
	      break;
	      
	    case 'L':
	      // scroll up one line
	      scrollup ();
	      state = 0;
	      break;

	      // e.g. dialog
	    case 'G': case '`':
	      xpos = state == 1 ? arg[0]-1 : 0;
		
	      state = 0;
	      break;
	      
	      // e.g. zile, nvi
	    case 'A': // Cursor Up
	      // TODO: review state encoding - state == 2 for the remaining as well?
	      ypos -= state == 2 ? arg[0] : 1;
	      
	      state = 0;
	      break;
	     
	    case 'B': case 'e': // Cursor Down
	      ypos += state == 1 ? arg[0] : 1;
		
	      state = 0;
	      break;
	      
	      // e.g. bash history - state 1
	    case 'C': case 'a': // Cursor Right
	      xpos += state == 1 ? arg[0] : 1;
	      
	      state = 0;
	      break;

	    case 'D': // Cursor Left
	      xpos -= state == 1 ? arg[0] : 1;
	      
	      state = 0;
	      break;
	      
	    case 'E': // Cursor next line
	      xpos = 0;
	      ypos += state == 1 ? arg[0] : 1;
	      
	      state = 0;
	      break;

	    case 'F': // Cursor previous line
	      xpos = 0;
	      ypos -= state == 1 ? arg[0] : 1;
	      
	      state = 0;
	      break;

	      // e.g. zile, nvi
	    case 'd':
	      ypos = state == 1 ? arg[0]-1 : 0;
	      
	      state = 0;
	      break;
	      
	      // e.g. vim - state 2
	    case 'c':
	      /*
		if (!vc->vc_par[0])
		respond_ID(tty);
	      */
	      state = 0;
	      break;
	      
	      // e.g. bash history - state 1
	    case 'P': // delete arg[0] chars to the right?
	      { 
		int i = SCR(xpos, ypos);
		int n = arg[0] > 1 ? arg[0] : 1;
		
		for (int x = xpos; x < xsize; ++x, ++i)
		  {
		    if (x + n < xsize) {
		      screen[i] = screen[i+n];
		      screenfg[i] = screenfg[i+n];
		      screenbg[i] = screenbg[i+n];
		      screenul[i] = screenul[i+n];
		    }
		    else
		      clear_one (i);
		  }
	      }
	      
	      state = 0;
	      break;
	      
	      // e.g. bash - state 1
	    case '@': // insert arg[0] chars or 1
	      { 
		int i = SCR(xsize-1, ypos);
		int n = arg[0] > 1 ? arg[0] : 1;
		
		for (int x = xsize-1; x > xpos; --x, --i)
		  {
		    if (x-n > xpos) {
		      screen[i] = screen[i-n];
		      screenfg[i] = screenfg[i-n];
		      screenbg[i] = screenbg[i-n];
		      screenul[i] = screenul[i-n];
		    }
		    else
		      clear_one (i);
		  }
	      }
	      state = 0;
	      break;
	      
	      // e.g. nvi, irssi (in screen), dialog
	    case 'M': // delete arg[0] - y down?
	      {
		// TODO: use scroll area?
		int n = arg[0] > 1 ? arg[0] : 1;
		
		for (int y = ypos; y < ysize; ++y)
		  {
		    if (y+n < ysize) {
		      int i = SCR(0,y);
		      int j = SCR(0,y+n);
		      for (int x = 0; x < xsize; ++x,++i,++j) {
			screen   [i] = screen   [j];
			screenfg [i] = screenfg [j];
			screenbg [i] = screenbg [j];
			screenul [i] = screenul [j];
		      }
		    }
		    else
		      clear_line (y);
		  }
	      }
	      state = 0;
	      break;
	      
	      case 's': // Save cursor
		save();
		state = 0;
		break;
		
	      case 'u': // Unsave cursor
		unsave();
		state = 0;
		break;
	      
	      // e.g. dialog based stuff
	    case 'X': // erase vpar partition
	      {
		int i = SCR(xpos,ypos);
		int c = SCR(xsize-1, ysize-1) - i;
		if (arg[0] < c)
		  c = arg[0];
		
		while (c--)
		  clear_one (i++);
	      }
	      state = 0;
	      break;
	      
	      // TODO: probably more
	      
	    default:
	      std::cerr << "Unknown code: '" << (char)wc
			<< "' (#" << (int)wc
			<< " - state " << state << ")" << std::endl;
	      state = 0;
	    }
	  continue;
        }

	// acs = Alternate Character Set.
	use_acs = 0;
	if (acs)
	{
            /* FIXME: If we are using a non-UTF-8 locale, the wcwidth
               call below will almost certainly fail.  We should have
               hardcoded results to fall back on in that case.  This
               will probably be fixed when I make this code drastically
               less dependent on mbrtowc and wctomb, which I really
               haven't figured out how to do yet.  They aren't really
               appropriate for a terminal emulator to be using!  */
std::cerr << "if (acs)" << std::endl;
	    switch (wc)
	    {
	    case 'q':
	      wc = 0x2500;
	      use_acs = 1;
	      break;
	    case 'j':
	      wc = 0x2518;
	      use_acs = 1;
	      break;
	    case 'x':
	      wc = 0x2502;
	      use_acs = 1;
	      break;
	    case 'a':
	      wc = 0x2591;
	      use_acs = 1;
	      break;
	    case 'm':
	      wc = 0x2514;
	      use_acs = 1;
	      break;
	    case 'l':
	      wc = 0x250c;
	      use_acs = 1;
	      break;
	    case 'k':
	      wc = 0x2510;
	      use_acs = 1;
	      break;
	    case 'u':
	      wc = 0x2524;
	      use_acs = 1;
	      break;
	    case 't':
	      wc = 0x251c;
	      use_acs = 1;
	      break;
	    }
	}

	/* At this point, if we can not decode a character because of ACS,
	   replace it with a space to minimize graphical corruption.  */
        if ((w = wcwidth (wc)) < 0)
        {
std::cerr << "cdwidth failed" << std::endl;
            if (use_acs)
            {
                wc = 0x20;
                w = wcwidth (wc);
            } else {
std::cerr << "if (!use_acs)" << std::endl;
                continue;
	    }
        }

        wctomb (0, 0);
        kk = wctomb (buf, wc);
        if (kk == -1)           /* impossible */
            continue;
#if 0
        {
            write (2, "p", 1);
        }
#endif

        f = rev ? bg : fg;
        b = rev ? fg : bg;

	//std::cerr << "here: wc: " << (char) wc << " w: " << w << std::endl;
        
        if (w > 0)
        {
            if (w <= xsize)
            {

                if (xpos + w > xsize)
                {
		    // clear_left ();

                    for (i = SCR (xpos, ypos); i < SCR (xsize, ypos); i++)
                    {
                        if(!is_clear (i))
                        {
                            screen[i] = ' ';
                            /* Use fg and bg rather than f and b - this is not
                               affected by reverse video. */
                            screenfg[i] = fg;
                            screenbg[i] = bg;
                            screenul[i] = 0;
                        }
                    }

                    xpos = 0;
                    cursor_down ();
                }

                i = SCR (xpos, ypos);
                screen[i] = wc;
                screenfg[i] = f;
                screenbg[i] = b;
                screenul[i] = ul;

                xp = xpos, yp = ypos;
                xpos += w;
            }
        }
        else
        {                       /* w == 0 */
            if (txp >= 0)
            {
	      xp = txp;
            }
            else
	      {
		xp = xpos, yp = ypos;
                xpos += 1;
            }
        }
    }
}

void Term::redraw ()
{
  int x, y, i;
  
  for (y = 0; y < ysize; ++y)
    for (x = 0; x < xsize; ++x)
      {
	i = SCR(x, y);
	if (screen[i])
	  {
	    put_char (x, y, screen[i],
		      screenfg[i], screenbg[i],
		      screenul[i]);
	  }
      }
  if (cur_visible)
    show_cursor (1);
}
