
// some more utility stuff

/* This first tries the modern Unix98 way of getting a pty, followed by the
 * old-fashioned BSD way in case that fails. */
int get_ptytty (int *xptyfd, int *xttyfd)
{
  char buf[16];
  int i, ptyfd, ttyfd;

  ptyfd = open ("/dev/ptmx", O_RDWR);
  if (ptyfd >= 0) {
    const char *slave = ptsname (ptyfd);
    if (slave) {
      if (grantpt (ptyfd) >= 0) {
        if (unlockpt (ptyfd) >= 0) {
          ttyfd = open (slave, O_RDWR);
          if (ttyfd >= 0) {
            *xptyfd = ptyfd, *xttyfd = ttyfd;
            return 0;
          }
        }
      }
    }
    close(ptyfd);
  }

  for (i = 0; i < 32; i++) {
    sprintf(buf, "/dev/pty%c%x", "pqrs"[i/16], i%16);
    ptyfd = open(buf, O_RDWR);
    if (ptyfd < 0) {
      sprintf(buf, "/dev/pty/m%d", i);
      ptyfd = open(buf, O_RDWR);
    }
    if (ptyfd >= 0) {
      sprintf(buf, "/dev/tty%c%x", "pqrs"[i/16], i%16);
      ttyfd = open(buf, O_RDWR);
      if (ttyfd < 0) {
        sprintf(buf, "/dev/pty/s%d", i);
        ttyfd = open(buf, O_RDWR);
      }
      if (ttyfd >= 0) {
	*xptyfd = ptyfd, *xttyfd = ttyfd;
	return 0;
      }
      close (ptyfd);
      return 1;
    }
  }
  return 1;
}

/* Probably I should use this as a signal handler */
void send_hangup (void)
{
  if (child_pid)
    kill (child_pid, SIGHUP);
}

void sigchld (int sig)
{
  int status;
  if (waitpid (child_pid, &status, WNOHANG)) {
    child_pid = 0;
    /* Reset ownership and permissions of ttyfd device? */
    tcsetattr (0, TCSAFLUSH, &ttysave);
    if (WIFEXITED (status))
      exit (WEXITSTATUS (status));
    if (WIFSIGNALED (status))
      exit (128 + WTERMSIG (status));
    if (WIFSTOPPED (status))
      exit (128 + WSTOPSIG (status));
    exit (status);
  }
  signal (SIGCHLD, sigchld);
}

void spawn_shell (int ptyfd, int ttyfd, const char *command)
{
  fflush (stdout);
  child_pid = fork ();
  if (child_pid) {
    /* Change ownership and permissions of ttyfd device! */
    signal (SIGCHLD, sigchld);
    return;
  }
  
  // we do not want to expose an own ID, rather be fully compatible
  setenv ("TERM", "linux", 1);
  
  close (ptyfd);

  setsid ();
  ioctl (ttyfd, TIOCSCTTY, (char *)0);
  dup2 (ttyfd, 0);
  dup2 (ttyfd, 1);
  dup2 (ttyfd, 2);
  if (ttyfd > 2)
    close(ttyfd);
  tcsetattr (0, TCSANOW, &ttysave);
  setgid (getgid());
  setuid (getuid());

  execl (command, command, NULL);
  exit (127);
}

void set_window_size (int ttyfd, int x, int y)
{
  struct winsize win;

  win.ws_row = y;
  win.ws_col = x;
  win.ws_xpixel = 0;
  win.ws_xpixel = 0;
  ioctl (ttyfd, TIOCSWINSZ, &win);
}
